@@ -1,32 +1,35 @@
-#!/usr/bin/python3
-"""Protococo.
+#!/usr/bin/env python3
+"""protococo.
 
 Usage:
-  protococo.py check  <message_name> [<message_hex_string> ...]
+  protococo check  <message_name> [<message_hex_string> ...]
                       [--cocofile=<file> --format=<option>] 
-                      [--verbose --decode]
-  protococo.py find   [<message_hex_string> ...]
+                      [--verbose --decode --decode-no-newlines]
+  protococo find   [<message_hex_string> ...]
                       [--cocofile=<file> --format=<option>]
                       [--dissect | --dissect-fields=<comma_separated_fields>]
-                      [--list --verbose --decode --long-name]
-  protococo.py create (<message_name> | --from-json=<json_file>)
+                      [--list --verbose --decode --decode-no-newlines --long-names]
+  protococo create (<message_name> | --from-json=<json_file>)
                       [--cocofile=<file>]
-  protococo.py json-recipe <message_names> ...
+  protococo json-recipe <message_names> ...
                       [--cocofile=<file>]
 
 Options:
-  -h --help             Show this screen.
-  --version             Show version.
-  --cocofile=<file>     Specify the protococo rules file [default: default.coco].
-  --verbose             Enable verbose output.
-  --format=<option>     Print message disection in different formats [default: oneline].
-                            Options: oneline, multiline.
-  --dissect             Include message field dissection in find results.
-  --list                Include a list of the most fitting messages in find results.
+  -h --help                 Show this screen.
+  --version                 Show version.
+  --cocofile=<file>         Specify the protococo rules file [default: default.coco].
+  --verbose                 Enable verbose output.
+  --format=<option>         Print message disection in different formats [default: compact].
+                                Options: oneline, multiline, compact.
+  --dissect                 Include message field dissection in find results.
+  --decode                  Decodes fields with encodedas parameters in message dissection
+  --decode-no-newlines      Replaces new lines in decoded fields of message dissections with \'\\n\' for a more compact output
+  --long-names              Prints the full mangled message names if a name mangling preprocess has been made during cocofile parsing
+  --list                    Include a list of the most fitting messages in find results.
   
 """
 
-__version__ = "0.1.0"
+__version__ = "0.2.0"
 
 from pprint import *
 from collections import OrderedDict
@@ -37,33 +40,68 @@ import json
 import copy
 from docopt import docopt
 
-# TODO: Esto pide a gritos como mínimo 3 clases: FieldRule (con byte_symbol, field_name y params), SpecialRule (con params), TitleRule (con message_name)
+
+TOKEN_DELIMITER = '$'
+COMMENT_DELIMITER = '#'
+
+def remove_comments(line):
+    if line.count(COMMENT_DELIMITER) == 0:
+        return line
+    
+    result = line.strip()
+    
+    if line.count(TOKEN_DELIMITER) == 0:
+        result = line[:line.index(COMMENT_DELIMITER)]
+    else:
+        lhs = line[:line.index(TOKEN_DELIMITER)].strip()
+        if COMMENT_DELIMITER in lhs:
+            open_parenthesis_count = 0
+            for i, c in enumerate(lhs):
+                if c=="(":
+                    open_parenthesis_count+=1
+                elif c==")":
+                    open_parenthesis_count-=1
+                if open_parenthesis_count < 0:
+                    raise RuntimeError("Unexpected ')'")
+                if c == COMMENT_DELIMITER and open_parenthesis_count == 0:
+                    result = line[:i]
+        else:
+            result = line[:line.index(COMMENT_DELIMITER)]
+    
+    return result.strip()
+
+def tokenize_line(line):
+    tokenized_rule = []
+        
+    if line.count(TOKEN_DELIMITER) == 0:
+        if (rule_is_title([line])):
+            tokenized_rule += [line]
+        else:
+            raise RuntimeError(f"Non-title rule without '{TOKEN_DELIMITER}' character")
+    else:
+    
+        #left hand side
+        lhs = line[:line.index(TOKEN_DELIMITER)].strip()
+        tokenized_rule.append(lhs)
+        #right hand side
+        rhs = line[line.strip().index(TOKEN_DELIMITER)+1:]
+        tokenized_rhs = [token.strip() for token in rhs.split(',')]
+        tokenized_rule += tokenized_rhs
+    
+    return tokenized_rule
 
 def tokenize_rules(rules_string):
     lines = [line.strip() for line in rules_string.splitlines() if len(line.strip()) > 0]
 
-    tokenized_lines = []
+    tokenized_rules = []
     for line in lines:
-        tokenized_line = []
-        
-        if line.count('#') == 0:
-            if (rule_is_title([line])):
-                tokenized_line += [line]
-            else:
-                raise RuntimeError("Non-title rule without '#' character")
-        else:
-        
-            #left hand side
-            lhs = line[:line.index('#')].strip()
-            tokenized_line.append(lhs)
-            #right hand side
-            rhs = line[line.strip().index('#')+1:]
-            tokenized_rhs = [token.strip() for token in rhs.split(',')]
-            tokenized_line += tokenized_rhs
-        
-        tokenized_lines.append(tokenized_line)
+        line = remove_comments(line)
+        if line == "":
+            continue
+        rule = tokenize_line(line)
+        tokenized_rules.append(rule)
     
-    return tokenized_lines
+    return tokenized_rules
 
 def rule_is_special(tokenized_rule):
     return tokenized_rule[0] == ""
@@ -83,7 +121,7 @@ def rule_is_title(tokenized_rule):
     #print(f"{len(title_with_brackets)=}, {tokenized_rule[0]=}, {tokenized_rule[-1]=}")
     if len(tokenized_rule) != 1 or title_with_brackets[0] != '[' or title_with_brackets[-1] != ']':
         return False
-    elif tokenized_rule.count('#') > 0 or tokenized_rule.count(',') > 0:
+    elif tokenized_rule.count(TOKEN_DELIMITER) > 0 or tokenized_rule.count(',') > 0:
         raise RuntimeError("Unexpected character in title rule={tokenized_rule}")
     else:
         return True
@@ -297,8 +335,11 @@ def preprocess_multiple_subtypeof(all_messages_rules_tokenized):
                     for k, mr in enumerate(all_messages_rules_tokenized):
                         for m, r in enumerate(mr):
                             if rule_is_subtypeof(r):
-                                if title_rule_get_name(message_rules[0]) in subtypeof_rule_get_parent(r).split():
-                                    r = ["", f"subtypeof {' '.join(new_message_mangled_names)}"]
+                                r_parents = subtypeof_rule_get_parent(r).split()
+                                common_parent = title_rule_get_name(message_rules[0])
+                                if common_parent in r_parents:
+                                    r_parents.remove(common_parent)
+                                    r = ["", f"subtypeof {' '.join(r_parents + new_message_mangled_names)}"]
                                     all_messages_rules_tokenized[k][m] = r
         j+=1
     
@@ -883,6 +924,10 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                     
                     if field_rule_is_encoded(rule):
                         decoded_result_dict[field_name_mangled] = f"{field_decode(rule, message_field_subtring)}"
+                elif current_length == 0:
+                    diff_dict.update({field_name_mangled: True})
+                    result_dict.update({field_name_mangled: ""})
+
                 else:
                     current_length = 0
                     
@@ -936,27 +981,88 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
 
 
 def split_multimessage_rules(multimessage_rules_string):
-    multimessage_rules_string_without_empty_lines = os.linesep.join([s.strip() for s in multimessage_rules_string.splitlines() if s]) 
+    lines = [clean_line for line in multimessage_rules_string.splitlines() if (clean_line := remove_comments(line)) != '']
+    multimessage_rules_string_without_empty_lines = os.linesep.join(lines)
     list_of_message_rules = re.split("(\[[^\[]*)", multimessage_rules_string_without_empty_lines)
     list_of_message_rules = list(filter(lambda x: x!= "", list_of_message_rules))
     
     return list_of_message_rules
 
+def get_short_message_name(long_message_name):
+    return re.sub(r'.*?([^\.]*)$', r'\1', long_message_name)
+
+def get_max_message_name_length(all_messages_rules_tokenized, long_names = False):
+    max_length = 0
+    if long_names:
+        for message_rules in all_messages_rules_tokenized:
+            max_length = max(max_length, len(title_rule_get_name(message_rules[0])))
+    else:
+        for message_rules in all_messages_rules_tokenized:
+            max_length = max(max_length, len(get_short_message_name(title_rule_get_name(message_rules[0]))))
+    return max_length
+
 class AnsiColors:
-    HEADER = '\033[95m'
+    PURPLE = '\033[95m'
     OKBLUE = '\033[94m'
-    OKCYAN = '\033[96m'
     OKGREEN = '\033[92m'
+    OKCYAN = '\033[96m'
     WARNING = '\033[93m'
     FAIL = '\033[91m'
+    FAIL2 = '\033[38;5;196m'
     ENDC = '\033[0m'
     BOLD = '\033[1m'
     UNDERLINE = '\033[4m'
+    UNDERLINE_OFF = '\033[24m'
+    
+def get_message_explanation_string_compact(validation_result, filter_fields = None, decode=False, no_newlines=False):
+    _, validation_result_dict, validation_diff_dict, __, validation_decoded_dict = validation_result
+    
+    result_string = ""
+    odd = 0
+    for k, v in validation_result_dict.items():
+        if filter_fields is not None and k not in filter_fields:
+            continue
+        
+        odd ^= 1
+        
+        field_complies = validation_diff_dict[k]
+        
+        k_adj, v_adj = k, v
+        
+        if decode == True and k in validation_decoded_dict.keys():
+            v_adj = validation_decoded_dict[k]
+        
+        fail_color = AnsiColors.FAIL if odd == 1 else AnsiColors.FAIL2
+        ok_color = AnsiColors.OKGREEN if odd == 1 else AnsiColors.OKCYAN
+        if not field_complies:
+            color = fail_color
+        else:
+            color = ok_color
+        
+        v_adj = AnsiColors.BOLD + color + v_adj + AnsiColors.ENDC
+            
+        if decode == True and k in validation_decoded_dict.keys():
+            if no_newlines:
+                v_adj = f"({v_adj})".replace("\r", "").replace("\n", f"{AnsiColors.PURPLE}\\n{AnsiColors.UNDERLINE_OFF + AnsiColors.BOLD + color}")
+            else:
+                v_adj = f"({v_adj})"
+            
+        if k_adj is not None:
+            result_string += f"{v_adj}"
+        else:   # Overflowing bytes field
+            result_string += f"|+{v_adj}"
+    
 
-def get_message_explanation_string_oneline(validation_result, filter_fields = None, decode=False):
+    
+    return result_string
+
+def get_message_explanation_string_oneline(validation_result, filter_fields = None, decode=False, no_newlines=False):
     
     _, validation_result_dict, validation_diff_dict, __, validation_decoded_dict = validation_result
     
+    fail_color = AnsiColors.FAIL
+    ok_color = AnsiColors.OKGREEN
+
     result_string = ""
     for k, v in validation_result_dict.items():
         if filter_fields is not None and k not in filter_fields:
@@ -970,12 +1076,17 @@ def get_message_explanation_string_oneline(validation_result, filter_fields = No
             v_adj = validation_decoded_dict[k]
         
         if not field_complies:
-            v_adj = AnsiColors.BOLD + AnsiColors.FAIL + v_adj + AnsiColors.ENDC
+            color = fail_color
         else:
-            v_adj = AnsiColors.BOLD + AnsiColors.OKGREEN + v_adj + AnsiColors.ENDC
+            color = ok_color
+        
+        v_adj = AnsiColors.BOLD + color + v_adj + AnsiColors.ENDC
             
         if decode == True and k in validation_decoded_dict.keys():
-            v_adj = f"({v_adj})"
+            if no_newlines:
+                v_adj = f"({v_adj})".replace("\r", "").replace("\n", f"{AnsiColors.PURPLE}\\n{AnsiColors.UNDERLINE_OFF + AnsiColors.BOLD + color}")
+            else:
+                v_adj = f"({v_adj})"
             
         if k_adj is not None:
             k_adj = AnsiColors.BOLD + k_adj + AnsiColors.ENDC
@@ -989,10 +1100,13 @@ def get_message_explanation_string_oneline(validation_result, filter_fields = No
     return result_string
 
 
-def get_message_explanation_string_multiline(validation_result, filter_fields = None, decode=False):
+def get_message_explanation_string_multiline(validation_result, filter_fields = None, decode=False, no_newlines=False):
     
     _, validation_result_dict, validation_diff_dict, __, validation_decoded_dict = validation_result
     
+    fail_color = AnsiColors.FAIL
+    ok_color = AnsiColors.OKGREEN
+
     result_string_field_names = ""
     result_string_field_values = ""
     for k, v in validation_result_dict.items():
@@ -1011,12 +1125,20 @@ def get_message_explanation_string_multiline(validation_result, filter_fields =
             v_adj = "+" + v
         
         lendiff = len(k_adj) - len(v_adj)
+
+        if decode == True and k in validation_decoded_dict.keys():
+            lendiff-=2  #To compensate the fact that we are adding 2 parenthesis
+            if no_newlines:
+                v_adj = f"({v_adj})".replace("\r", "").replace("\n", f"{AnsiColors.PURPLE}\\n{AnsiColors.UNDERLINE_OFF + AnsiColors.BOLD + color}")
+            else:
+                v_adj = f"({v_adj})"
         
         if not field_complies:
-            v_adj = AnsiColors.BOLD + AnsiColors.FAIL + v_adj + AnsiColors.ENDC
+            color = fail_color
         else:
-            v_adj = AnsiColors.BOLD + AnsiColors.OKGREEN + v_adj + AnsiColors.ENDC
-        k_adj = AnsiColors.BOLD + k_adj + AnsiColors.ENDC
+            color = ok_color
+        
+        v_adj = AnsiColors.BOLD + color + v_adj + AnsiColors.ENDC
             
         if lendiff < 0:
             prefix = " " * ((-lendiff)//2)
@@ -1041,14 +1163,16 @@ def get_message_explanation_string_multiline(validation_result, filter_fields =
     return result_string_field_names + "\n" + result_string_field_values
 
 
-def get_message_explanation_string(validation_result, validation_log_dict = None, oneline = False, filter_fields = None, decode = False):
+def get_message_explanation_string(validation_result, validation_log_dict = None, fmt="oneline", filter_fields = None, decode = False, no_newlines=False):
     
     _, validation_result_dict, validation_diff_dict, __, ___ = validation_result
     
-    if oneline == True:
-        result_string = get_message_explanation_string_oneline(validation_result, filter_fields, decode=decode)
+    if fmt == "oneline":
+        result_string = get_message_explanation_string_oneline(validation_result, filter_fields, decode=decode, no_newlines=no_newlines)
+    elif fmt == "compact":
+        result_string = get_message_explanation_string_compact(validation_result, filter_fields, decode=decode, no_newlines=no_newlines)
     else:
-        result_string = get_message_explanation_string_multiline(validation_result, filter_fields, decode=decode)
+        result_string = get_message_explanation_string_multiline(validation_result, filter_fields, decode=decode, no_newlines=no_newlines)
 
     logs_string = ""
     if validation_log_dict is not None and len(validation_log_dict) > 0:
@@ -1284,6 +1408,7 @@ def cli_main():
     
     ret = 0
         
+    
     all_messages_rules_tokenized = [tokenize_rules(r) for r in split_multimessage_rules(all_messages_string)]
     preprocess_encode_fields(all_messages_rules_tokenized)
     all_messages_rules_tokenized = preprocess_multiple_subtypeof(all_messages_rules_tokenized)
@@ -1298,16 +1423,13 @@ def cli_main():
             if args["--verbose"] == True:
                 explanation_logs = validate_result[3]
                 
-            print(get_message_explanation_string(validate_result, explanation_logs, args["--format"] == "oneline", decode=args["--decode"]))
+            print(get_message_explanation_string(validate_result, explanation_logs, fmt=args["--format"], decode=args["--decode"], no_newlines=args["--decode-no-newlines"]))
             
             if validate_result[0] == False:
                 ret = 1
         
     elif args["find"] == True:
         messages_input = sys.stdin.read().split() if not args["<message_hex_string>"] else args["<message_hex_string>"]
-        
-        oneline_enabled = args["--format"] == "oneline"
-        
         for message_hex_string in messages_input:
             ordered_message_names, validate_results_by_message_name = identify_message(message_hex_string, all_messages_rules_tokenized)
             for i, match in enumerate(ordered_message_names):
@@ -1320,28 +1442,29 @@ def cli_main():
                     validate_result = validate_message_by_name(match, message_hex_string, all_messages_rules_tokenized)
                     
                     if args["--verbose"] == True:
-                        explanation = "\n" + get_message_explanation_string(validate_result, validate_result[3], oneline_enabled, filter_fields=filter_fields, decode=args["--decode"])
+                        explanation = "\n" + get_message_explanation_string(validate_result, validate_result[3], fmt=args["--format"], filter_fields=filter_fields, decode=args["--decode"], no_newlines=args["--decode-no-newlines"])
                     else:
-                        explanation = get_message_explanation_string(validate_result, None, oneline_enabled, filter_fields=filter_fields, decode=args["--decode"])
+                        explanation = get_message_explanation_string(validate_result, None, fmt=args["--format"], filter_fields=filter_fields, decode=args["--decode"], no_newlines=args["--decode-no-newlines"])
                         
                     if validate_result[0] == False:
                         ret = 1
                 
                 name_string = match
-                if args["--long-name"] == False:
-                    name_string = re.sub(r'.*?([^\.]*)$', r'\1', match)
+                if args["--long-names"] == False:
+                    name_string = get_short_message_name(match)
+                number_of_whitespaces = get_max_message_name_length(all_messages_rules_tokenized, args["--long-names"]) - len(name_string) + 2
                     
                 if args["--list"] == False:
-                    if oneline_enabled:
-                        print(color  + f"[{name_string}]" + AnsiColors.ENDC + "\t" + explanation)
+                    if args["--format"] == "oneline" or args["--format"] == "compact":
+                        print(color  + f"[{name_string}]" + AnsiColors.ENDC + " "*number_of_whitespaces + explanation)
                     else:
                         print(color  + f"[{name_string}]" + AnsiColors.ENDC)
                         print(explanation)
                         print()
                     break
                 else:
-                    if oneline_enabled:
-                        print(color  + f"- {i}: [{name_string}]" + AnsiColors.ENDC + "\t" + explanation)
+                    if args["--format"] == "oneline" or args["--format"] == "compact":
+                        print(color  + f"{str(i): >8}: [{name_string}]" + AnsiColors.ENDC + " "*number_of_whitespaces + explanation)
                     else:
                         print(color  + f"- {i}: [{name_string}]" + AnsiColors.ENDC)
                         print(explanation)
@@ -1417,6 +1540,10 @@ def cli_main():
 #TODO optimization: if a parent rule fails, don't check subtypes. --list'd not be possible
 #TODO refactor: CocoDocument, CocoMessageSpec, CocoRule, CocoParser, CocoAnalyzer, CocoCLI
 #TODO improvement: cocofile checks: no "." in any rule
+#TODO fix: problem decoding littleendian from rule between parenthesis, example: (0)   #encodedas littlendian
+#TODO fix: override a 4 byte field with a 1 byte field. example: (0)   #encodedas littlendian with parent like XXXXXXXX
+#TODO feature: add encodedas json
+#TODO fix?: override field from different parent levels
 
             
 if __name__ == "__main__":
